"""\
Various useful plugins for working with bzr.
"""
import bzrlib.commands
import push
import annotate
import shelf
import sys
import os.path
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__), 
						 "external")))
from bzrlib.option import Option

Option.OPTIONS['ignored'] = Option('ignored',
        help='delete all ignored files.')
Option.OPTIONS['detrius'] = Option('detrius',
        help='delete conflict files merge backups, and failed selftest dirs.' +
              '(*.THIS, *.BASE, *.OTHER, *~, *.tmp')
Option.OPTIONS['dry-run'] = Option('dry-run',
        help='show files to delete instead of deleting them.')

class cmd_clean_tree(bzrlib.commands.Command):
    """Remove unwanted files from working tree.
    Normally, ignored files are left alone.
    """
    takes_options = ['ignored', 'detrius', 'dry-run']
    def run(self, ignored=False, detrius=False, dry_run=False):
        from clean_tree import clean_tree
        clean_tree('.', ignored=ignored, detrius=detrius, dry_run=dry_run)

Option.OPTIONS['no-collapse'] = Option('no-collapse')
Option.OPTIONS['no-antialias'] = Option('no-antialias')
Option.OPTIONS['cluster'] = Option('cluster')
Option.OPTIONS['merge-branch'] = Option('merge-branch',type=str)

class cmd_graph_ancestry(bzrlib.commands.Command):
    """Produce ancestry graphs using dot.
    
    Output format is detected according to file extension.  Some of the more
    common output formats are png, gif, svg, ps.  An extension of '.dot' will
    cause a dot graph file to be produced.

    Branches are labeled r?, where ? is the revno.  If they have no revno,
    with the last 5 characters of their revision identifier are used instead.
    
    If --merge-branch is specified, the two branches are compared and a merge
    base is selected.
    
    Legend:
    white    normal revision
    yellow   THIS  history
    red      OTHER history
    orange   COMMON history
    blue     COMMON non-history ancestor
    dotted   Missing from branch storage

    Ancestry is usually collapsed by removing revisions with a single parent
    and descendant.  The number of skipped revisions is shown on the arrow.
    This feature can be disabled with --no-collapse.

    By default, revisions are ordered by distance from root, but they can be
    clustered instead using --cluster.

    If available, rsvg is used to antialias PNG and JPEG output, but this can
    be disabled with --no-antialias.
    """
    takes_args = ['branch', 'file']
    takes_options = ['no-collapse', 'no-antialias', 'merge-branch', 'cluster']
    def run(self, branch, file, no_collapse=False, no_antialias=False,
        merge_branch=None, cluster=False):
        import graph
        if cluster:
            ranking = "cluster"
        else:
            ranking = "forced"
        graph.write_ancestry_file(branch, file, not no_collapse, 
                                  not no_antialias, merge_branch, ranking)

class cmd_fetch_ghosts(bzrlib.commands.Command):
    """Attempt to retrieve ghosts from another branch.
    If the other branch is not supplied, the last-pulled branch is used.
    """
    aliases = ['fetch-missing']
    takes_args = ['branch?']
    def run(self, branch=None):
        from fetch_ghosts import fetch_ghosts
        fetch_ghosts(branch)

class cmd_patch(bzrlib.commands.Command):
    """Apply a named patch to the current tree.
    """
    takes_args = ['filename?']
    takes_options = ['strip']
    def run(self, filename=None, strip=0):
        from patch import patch
        from bzrlib.branch import Branch
        b = Branch.open_containing('.')[0]
        return patch(b, filename, strip)



commands = [push.cmd_push, shelf.cmd_shelve, 
            shelf.cmd_unshelve, cmd_clean_tree, cmd_graph_ancestry,
            cmd_fetch_ghosts, cmd_patch]

import bzrlib.builtins
if not hasattr(bzrlib.builtins, "cmd_annotate"):
    commands.append(annotate.cmd_annotate)

from errors import NoPyBaz
try:
    import baz_import
    commands.append(baz_import.cmd_baz_import_branch)
    commands.append(baz_import.cmd_baz_import)

except NoPyBaz:
    class cmd_baz_import(bzrlib.commands.Command):
        """Disabled. (Requires PyBaz)"""
        takes_args = ['to_root_dir?', 'from_archive?']
        takes_options = ['verbose']
        def run(self, to_root_dir=None, from_archive=None, verbose=False):
            print "This command is disabled.  Please install PyBaz."
    commands.append(cmd_baz_import)

if hasattr(bzrlib.commands, 'register_command'):
    for command in commands:
        bzrlib.commands.register_command(command)

def test_suite():
    import baz_import
    import tests
    from doctest import DocTestSuite
    from unittest import TestSuite
    import clean_tree
    result = TestSuite()
    result.addTest(DocTestSuite(bzrtools))
    result.addTest(clean_tree.test_suite())
    result.addTest(DocTestSuite(baz_import))
    result.addTest(tests.test_suite())
    return result
